@@ -4,11 +4,30 @@ import abc
 import collections.abc
 import enum
 import typing
-from . import embedded
-from . import memorymanagement
-from . import openrlib
-from . import _rinterface_capi as _rinterface
-from . import conversion
+from rpy2.rinterface_lib import embedded
+from rpy2.rinterface_lib import memorymanagement
+from rpy2.rinterface_lib import openrlib
+import rpy2.rinterface_lib._rinterface_capi as _rinterface
+from rpy2.rinterface_lib._rinterface_capi import _evaluated_promise
+from rpy2.rinterface_lib._rinterface_capi import SupportsSEXP
+from rpy2.rinterface_lib import conversion
+from rpy2.rinterface_lib.conversion import _cdata_res_to_rinterface
+from rpy2.rinterface_lib import na_values
+
+
+class Singleton(type):
+
+    _instances: typing.Dict[typing.Type['Singleton'], 'Singleton'] = {}
+
+    def __call__(cls, *args, **kwargs):
+        instances = cls._instances
+        if cls not in instances:
+            instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
+        return instances[cls]
+
+
+class SingletonABC(Singleton, abc.ABCMeta):
+    pass
 
 
 class RTYPES(enum.IntEnum):
@@ -45,7 +64,7 @@ class RTYPES(enum.IntEnum):
     FUNSXP = openrlib.rlib.FUNSXP
 
 
-class Sexp(object):
+class Sexp(SupportsSEXP):
     """Base class for R objects.
 
     The name of a class corresponds to the name SEXP
@@ -53,14 +72,13 @@ class Sexp(object):
 
     __slots__ = ('_sexpobject', )
 
-    def __init__(self, sexp: typing.Union['Sexp', '_rinterface.SexpCapsule']):
-        if not embedded.isready():
-            raise embedded.RNotReadyError('Embedded R is not ready to use.')
-
-        if isinstance(sexp, Sexp):
+    def __init__(self,
+                 sexp: typing.Union[SupportsSEXP,
+                                    '_rinterface.SexpCapsule',
+                                    '_rinterface.UninitializedRCapsule']):
+        if isinstance(sexp, SupportsSEXP):
             self._sexpobject = sexp.__sexp__
-        elif (isinstance(sexp, _rinterface.SexpCapsule) or
-              isinstance(sexp, _rinterface.UnmanagedSexpCapsule)):
+        elif isinstance(sexp, _rinterface.CapsuleBase):
             self._sexpobject = sexp
         else:
             raise ValueError(
@@ -73,7 +91,7 @@ class Sexp(object):
         return super().__repr__() + (' [%s]' % self.typeof)
 
     @property
-    def __sexp__(self) -> '_rinterface.SexpCapsule':
+    def __sexp__(self) -> '_rinterface.CapsuleBase':
         """Access to the underlying C pointer to the R object.
 
         When assigning a new SexpCapsule to this attribute, the
@@ -83,7 +101,7 @@ class Sexp(object):
 
     @__sexp__.setter
     def __sexp__(self,
-                 value: '_rinterface.SexpCapsule') -> None:
+                 value: '_rinterface.CapsuleBase') -> None:
         assert isinstance(value, _rinterface.SexpCapsule)
         if value.typeof != self.__sexp__.typeof:
             raise ValueError('New capsule type mismatch: %s' %
@@ -101,8 +119,9 @@ class Sexp(object):
     def __getstate__(self) -> bytes:
         with memorymanagement.rmemory() as rmemory:
             ser = rmemory.protect(
-                _rinterface.serialize(self.__sexp__._cdata,
-                                      embedded.globalenv.__sexp__._cdata)
+                _rinterface.serialize(
+                    self.__sexp__._cdata,
+                    globalenv.__sexp__._cdata)
             )
             n = openrlib.rlib.Rf_xlength(ser)
             res = bytes(_rinterface.ffi.buffer(openrlib.rlib.RAW(ser), n))
@@ -176,7 +195,7 @@ class Sexp(object):
 
     @property
     def names(self) -> 'Sexp':
-        return embedded.baseenv['names'](self)
+        return baseenv['names'](self)
 
     @names.setter
     def names(self, value) -> None:
@@ -193,6 +212,33 @@ class Sexp(object):
             openrlib.rlib.R_NameSymbol)
 
 
+class NULLType(Sexp, metaclass=SingletonABC):
+    """A singleton class for R's NULL."""
+
+    def __init__(self):
+        if embedded.isready():
+            tmp = Sexp(
+                _rinterface.UnmanagedSexpCapsule(
+                    openrlib.rlib.R_NilValue
+                )
+            )
+        else:
+            tmp = Sexp(_rinterface.UninitializedRCapsule(RTYPES.NILSXP.value))
+        super().__init__(tmp)
+
+    def __bool__(self) -> bool:
+        """This is always False."""
+        return False
+
+    @property
+    def __sexp__(self) -> _rinterface.CapsuleBase:
+        return self._sexpobject
+
+    @property
+    def rid(self) -> int:
+        return self._sexpobject.rid
+
+
 # TODO: Duplicate declaration in R_API.h ?
 class CETYPE(enum.Enum):
     """Character encodings for R string."""
@@ -232,6 +278,184 @@ class CharSexp(Sexp):
                                      self._NCHAR_MSG)
 
 
+class SexpEnvironment(Sexp):
+    """Proxy for an R "environment" object.
+
+    An R "environment" object can be thought of as a mix of a
+    mapping (like a `dict`) and a scope. To make it more "Pythonic",
+    both aspects are kept separate and the method `__getitem__` will
+    get an item as it would for a Python `dict` while the method `find`
+    will get an item as if it was a scope.
+
+    As soon as R is initialized the following main environments become
+    available to the user:
+    - `globalenv`: The "workspace" for the current R process. This can
+      be thought of as when `__name__ == '__main__'` in Python.
+    - `baseenv`: The namespace of R's "base" package.
+    """
+
+    @_cdata_res_to_rinterface
+    @_evaluated_promise
+    def find(self,
+             key: str,
+             wantfun: bool = False) -> Sexp:
+        """Find an item, starting with this R environment.
+
+        Raises a `KeyError` if the key cannot be found.
+
+        This method is called `find` because it is somewhat different
+        from the method :meth:`get` in Python mappings such :class:`dict`.
+        This is looking for a key across enclosing environments, returning
+        the first key found."""
+
+        if not isinstance(key, str):
+            raise TypeError('The key must be a non-empty string.')
+        elif not len(key):
+            raise ValueError('The key must be a non-empty string.')
+        with memorymanagement.rmemory() as rmemory:
+            symbol = rmemory.protect(
+                openrlib.rlib.Rf_install(conversion._str_to_cchar(key))
+            )
+            if wantfun:
+                # One would expect this to be like
+                #   res = _rinterface._findfun(symbol, self.__sexp__._cdata)
+                # but R's findfun will segfault if the symbol is not in
+                # the environment. :/
+                rho = self
+                while rho.rid != emptyenv.rid:
+                    res = _rinterface._findVarInFrame(symbol,
+                                                      rho.__sexp__._cdata)
+                    if _rinterface._TYPEOF(res) in (openrlib.rlib.CLOSXP,
+                                                    openrlib.rlib.BUILTINSXP):
+                        break
+                    # TODO: move check of R_UnboundValue to _rinterface ?
+                    res = openrlib.rlib.R_UnboundValue
+                    rho = rho.enclos
+            else:
+                res = _rinterface._findvar(symbol, self.__sexp__._cdata)
+        # TODO: move check of R_UnboundValue to _rinterface ?
+        if res == openrlib.rlib.R_UnboundValue:
+            raise KeyError("'%s' not found" % key)
+        return res
+
+    @_cdata_res_to_rinterface
+    @_evaluated_promise
+    def __getitem__(self, key: str) -> typing.Any:
+        if not isinstance(key, str):
+            raise TypeError('The key must be a non-empty string.')
+        elif not len(key):
+            raise ValueError('The key must be a non-empty string.')
+        with memorymanagement.rmemory() as rmemory:
+            symbol = rmemory.protect(
+                openrlib.rlib.Rf_install(conversion._str_to_cchar(key))
+            )
+            res = _rinterface._findVarInFrame(symbol, self.__sexp__._cdata)
+        # TODO: move check of R_UnboundValue to _rinterface
+        if res == openrlib.rlib.R_UnboundValue:
+            raise KeyError("'%s' not found" % key)
+        return res
+
+    def __setitem__(self, key: str, value) -> None:
+        # TODO: move body to _rinterface-level function
+        if not isinstance(key, str):
+            raise TypeError('The key must be a non-empty string.')
+        elif not len(key):
+            raise ValueError('The key must be a non-empty string.')
+        if (self.__sexp__._cdata == openrlib.rlib.R_BaseEnv) or \
+           (self.__sexp__._cdata == openrlib.rlib.R_EmptyEnv):
+            raise ValueError('Cannot remove variables from the base or '
+                             'empty environments.')
+        # TODO: call to Rf_duplicate needed ?
+        with memorymanagement.rmemory() as rmemory:
+            symbol = rmemory.protect(
+                openrlib.rlib.Rf_install(conversion._str_to_cchar(key))
+            )
+            cdata = rmemory.protect(conversion._get_cdata(value))
+            cdata_copy = rmemory.protect(
+                openrlib.rlib.Rf_duplicate(cdata)
+            )
+            openrlib.rlib.Rf_defineVar(symbol,
+                                       cdata_copy,
+                                       self.__sexp__._cdata)
+
+    def __len__(self) -> int:
+        with memorymanagement.rmemory() as rmemory:
+            symbols = rmemory.protect(
+                openrlib.rlib.R_lsInternal(self.__sexp__._cdata,
+                                           openrlib.rlib.TRUE)
+            )
+            n = openrlib.rlib.Rf_xlength(symbols)
+        return n
+
+    def __delitem__(self, key: str) -> None:
+        # Testing that key is a non-empty string is implicitly
+        # performed when checking that the key is in the environment.
+        if key not in self:
+            raise KeyError("'%s' not found" % key)
+
+        if self.__sexp__ == baseenv.__sexp__:
+            raise ValueError('Values from the R base environment '
+                             'cannot be removed.')
+        # TODO: also check it is not R_EmpyEnv or R_BaseNamespace
+        if self.is_locked():
+            ValueError('Cannot remove an item from a locked '
+                       'environment.')
+
+        with memorymanagement.rmemory() as rmemory:
+            key_cdata = rmemory.protect(
+                openrlib.rlib.Rf_mkString(conversion._str_to_cchar(key))
+            )
+            _rinterface._remove(key_cdata,
+                                self.__sexp__._cdata,
+                                openrlib.rlib.Rf_ScalarLogical(
+                                    openrlib.rlib.FALSE))
+
+    @_cdata_res_to_rinterface
+    def frame(self) -> 'typing.Union[NULLType, SexpEnvironment]':
+        """Get the parent frame of the environment."""
+        return openrlib.rlib.FRAME(self.__sexp__._cdata)
+
+    @property
+    @_cdata_res_to_rinterface
+    def enclos(self) -> 'typing.Union[NULLType, SexpEnvironment]':
+        """Get or set the enclosing environment."""
+        return openrlib.rlib.ENCLOS(self.__sexp__._cdata)
+
+    @enclos.setter
+    def enclos(self, value: 'SexpEnvironment') -> None:
+        assert isinstance(value, SexpEnvironment)
+        openrlib.rlib.SET_ENCLOS(self.__sexp__._cdata,
+                                 value.__sexp__._cdata)
+
+    def keys(self) -> typing.Generator[str, None, None]:
+        """Generator over the keys (symbols) in the environment."""
+        with memorymanagement.rmemory() as rmemory:
+            symbols = rmemory.protect(
+                openrlib.rlib.R_lsInternal(self.__sexp__._cdata,
+                                           openrlib.rlib.TRUE)
+            )
+            n = openrlib.rlib.Rf_xlength(symbols)
+            res = []
+            for i in range(n):
+                res.append(_rinterface._string_getitem(symbols, i))
+        for e in res:
+            yield e
+
+    def __iter__(self) -> typing.Generator[str, None, None]:
+        """See method `keys()`."""
+        return self.keys()
+
+    def is_locked(self) -> bool:
+        return openrlib.rlib.R_EnvironmentIsLocked(
+            self.__sexp__._cdata)
+
+
+_UNINIT_CAPSULE_ENV = _rinterface.UninitializedRCapsule(RTYPES.ENVSXP.value)
+emptyenv = SexpEnvironment(_UNINIT_CAPSULE_ENV)
+baseenv = SexpEnvironment(_UNINIT_CAPSULE_ENV)
+globalenv = SexpEnvironment(_UNINIT_CAPSULE_ENV)
+
+
 # TODO: move to _rinterface-level function (as ABI / API compatibility
 # will have API-defined code compile for efficiency).
 def _populate_r_vector(iterable, r_vector, set_elt, cast_value):
@@ -285,7 +509,7 @@ class SexpVector(Sexp, metaclass=abc.ABCMeta):
                                                _rinterface.SexpCapsule):
             super().__init__(obj)
         elif isinstance(obj, collections.abc.Sized):
-            super().__init__(type(self).from_object(obj).__sexp__)
+            super().__init__(self.from_object(obj).__sexp__)
         else:
             raise TypeError('The constructor must be called '
                             'with an instance of '
@@ -294,35 +518,31 @@ class SexpVector(Sexp, metaclass=abc.ABCMeta):
                             'rpy2.rinterface._rinterface.SexpCapsule')
 
     @classmethod
-    def _populate_r_vector(cls, iterable,
-                           r_vector):
-        return _populate_r_vector(iterable,
-                                  r_vector,
-                                  cls._R_SET_VECTOR_ELT,
-                                  cls._CAST_IN)
-
-    @classmethod
-    @conversion._cdata_res_to_rinterface
+    @_cdata_res_to_rinterface
     def from_iterable(cls, iterable,
-                      populate_func=None) -> VT:
+                      populate_func=None,
+                      set_elt=None,
+                      cast_value=None) -> VT:
         """Create an R vector/array from an iterable."""
         if not embedded.isready():
             raise embedded.RNotReadyError('Embedded R is not ready to use.')
+        if populate_func is None:
+            populate_func = _populate_r_vector
+        if set_elt is None:
+            set_elt = cls._R_SET_VECTOR_ELT
+        if cast_value is None:
+            cast_value = cls._CAST_IN
         n = len(iterable)
         with memorymanagement.rmemory() as rmemory:
             r_vector = rmemory.protect(
                 openrlib.rlib.Rf_allocVector(
                     cls._R_TYPE, n)
             )
-            if populate_func is None:
-                cls._populate_r_vector(iterable,
-                                       r_vector)
-            else:
-                populate_func(iterable, r_vector)
+            populate_func(iterable, r_vector, set_elt, cast_value)
         return r_vector
 
     @classmethod
-    @conversion._cdata_res_to_rinterface
+    @_cdata_res_to_rinterface
     def from_memoryview(cls, mview: memoryview) -> VT:
         """Create an R vector/array from a memoryview.
 
@@ -380,21 +600,20 @@ class SexpVector(Sexp, metaclass=abc.ABCMeta):
 
     def __getitem__(
             self,
-            i: typing.Union[int, slice]) -> typing.Union[Sexp, VT]:
+            i: typing.Union[int, slice]) -> typing.Union[Sexp, VT, typing.Any]:
         cdata = self.__sexp__._cdata
         if isinstance(i, int):
             i_c = _rinterface._python_index_to_c(cdata, i)
             res = conversion._cdata_to_rinterface(
                 self._R_VECTOR_ELT(cdata, i_c))
         elif isinstance(i, slice):
-            res = type(self).from_iterable(
+            res = self.from_iterable(
                 [
                     self._R_VECTOR_ELT(
-                        cdata, i_c
+                        cdata, i_c,
                     ) for i_c in range(*i.indices(len(self)))
                 ],
-                populate_func=lambda iterable, r_vector: _populate_r_vector(
-                    iterable, r_vector, self._R_SET_VECTOR_ELT, lambda x: x)
+                cast_value=lambda x: x
             )
         else:
             raise TypeError(
@@ -418,7 +637,11 @@ class SexpVector(Sexp, metaclass=abc.ABCMeta):
     def __len__(self) -> int:
         return openrlib.rlib.Rf_xlength(self.__sexp__._cdata)
 
-    def index(self, item) -> int:
+    def __iter__(self) -> typing.Iterator[typing.Union[Sexp, VT, typing.Any]]:
+        for i in range(len(self)):
+            yield self[i]
+
+    def index(self, item: typing.Any) -> int:
         for i, e in enumerate(self):
             if e == item:
                 return i
@@ -444,13 +667,16 @@ class StrSexpVector(SexpVector):
 
     def __getitem__(
             self,
-            i: typing.Union[int, slice]) -> 'typing.Union[str, StrSexpVector]':
+            i: typing.Union[int, slice]
+    ) -> typing.Union['StrSexpVector', str, 'na_values.NA_Character']:
         cdata = self.__sexp__._cdata
         if isinstance(i, int):
             i_c = _rinterface._python_index_to_c(cdata, i)
             res = _rinterface._string_getitem(cdata, i_c)
+            if res is None:
+                res = na_values.NA_Character
         elif isinstance(i, slice):
-            res = type(self).from_iterable(
+            res = self.from_iterable(
                 [_rinterface._string_getitem(cdata, i_c)
                  for i_c in range(*i.indices(len(self)))]
             )
@@ -459,21 +685,36 @@ class StrSexpVector(SexpVector):
                             ' not %s' % type(i))
         return res
 
-    def __setitem__(self,
-                    i: typing.Union[int, slice],
-                    value: typing.Union[str, typing.Sequence[str]]) -> None:
+    def __setitem__(
+            self,
+            i: typing.Union[int, slice],
+            value: typing.Union[str, typing.Sequence[typing.Optional[str]],
+                                'StrSexpVector', 'na_values.NA_Character']
+    ) -> None:
         cdata = self.__sexp__._cdata
         if isinstance(i, int):
             i_c = _rinterface._python_index_to_c(cdata, i)
+            if isinstance(value, Sexp):
+                val_cdata = value.__sexp__._cdata
+            else:
+                if not isinstance(value, str):
+                    value = str(value)
+                val_cdata = _as_charsxp_cdata(value)
             self._R_SET_VECTOR_ELT(
                 cdata, i_c,
-                _as_charsxp_cdata(value)
+                val_cdata
             )
         elif isinstance(i, slice):
             for i_c, v in zip(range(*i.indices(len(self))), value):
+                if v is None:
+                    v_cdata = openrlib.rlib.R_NaString
+                else:
+                    if not isinstance(value, str):
+                        v = str(v)
+                    v_cdata = _as_charsxp_cdata(v)
                 self._R_SET_VECTOR_ELT(
                     cdata, i_c,
-                    _as_charsxp_cdata(v)
+                    v_cdata
                 )
         else:
             raise TypeError('Indices must be integers or slices, '
@@ -502,7 +743,7 @@ _DEFAULT_RCLASS_NAMES = {
     RTYPES.LANGSXP: 'language'}
 
 
-def rclass_get(scaps: _rinterface.SexpCapsule) -> StrSexpVector:
+def rclass_get(scaps: _rinterface.CapsuleBase) -> StrSexpVector:
     rlib = openrlib.rlib
     with memorymanagement.rmemory() as rmemory:
         classes = rmemory.protect(
@@ -530,7 +771,7 @@ def rclass_get(scaps: _rinterface.SexpCapsule) -> StrSexpVector:
 
 
 def rclass_set(
-        scaps: _rinterface.SexpCapsule,
+        scaps: _rinterface.CapsuleBase,
         value: 'typing.Union[StrSexpVector, str]'
 ) -> None:
     if isinstance(value, StrSexpVector):
@@ -555,7 +796,93 @@ def unserialize(state):
             openrlib.rlib.RAW(cdata), state, n)
         ser = rmemory.protect(
             _rinterface.unserialize(cdata,
-                                    embedded.globalenv.__sexp__._cdata)
+                                    globalenv.__sexp__._cdata)
         )
         res = _rinterface.SexpCapsule(ser)
     return res
+
+
+class NAIntegerType(int, metaclass=Singleton):
+
+    def __new__(cls, *args, **kwargs):
+        embedded.assert_isready()
+        return super().__new__(cls, openrlib.rlib.R_NaInt)
+
+    def __repr__(self) -> str:
+        return 'NA_integer_'
+
+    def __str__(self) -> str:
+        return 'NA_integer_'
+
+    def __bool__(self):
+        raise ValueError('R value for missing integer value')
+
+
+class NACharacterType(CharSexp, metaclass=SingletonABC):
+
+    def __init__(self):
+        embedded.assert_isready()
+        super().__init__(
+            CharSexp(
+                _rinterface.SexpCapsule(openrlib.rlib.R_NaString)
+            )
+        )
+
+    def __repr__(self) -> str:
+        return 'NA_character_'
+
+    def __str__(self) -> str:
+        return 'NA_character_'
+
+    def __bool__(self):
+        raise ValueError('R value for missing character value')
+
+
+class NALogicalType(int, metaclass=Singleton):
+
+    def __new__(cls, *args, **kwargs):
+        embedded.assert_isready()
+        return super().__new__(cls, openrlib.rlib.R_NaInt)
+
+    def __repr__(self) -> str:
+        return 'NA'
+
+    def __str__(self) -> str:
+        return 'NA'
+
+    def __bool__(self) -> bool:
+        raise ValueError('R value for missing boolean value')
+
+
+class NARealType(float, metaclass=Singleton):
+
+    def __new__(cls, *args, **kwargs):
+        embedded.assert_isready()
+        return super().__new__(cls, openrlib.rlib.R_NaReal)
+
+    def __repr__(self) -> str:
+        return 'NA_real_'
+
+    def __str__(self) -> str:
+        return 'NA_real_'
+
+    def __bool__(self) -> bool:
+        raise ValueError('R value for missing float value')
+
+
+class NAComplexType(complex, metaclass=Singleton):
+
+    def __new__(cls, *args, **kwargs):
+        embedded.assert_isready()
+        return super().__new__(cls,
+                               openrlib.rlib.R_NaReal,
+                               openrlib.rlib.R_NaReal)
+
+    def __repr__(self) -> str:
+        return 'NA_complex_'
+
+    def __str__(self) -> str:
+        return 'NA_complex_'
+
+    def __bool__(self):
+        raise ValueError('R value for missing complex value')
